;;;; crunch - embedding macro

(module crunch (crunch : crunch-options
                string-result u8vector-result s8vector-result
                u16vector-result s16vector-result u32vector-result
                s32vector-result u64vector-result s64vector-result
                intvector-result f32vector-result f64vector-result
                c64vector-result c128vector-result)

  (import (scheme base)
          (chicken base)
          (chicken syntax)
          (chicken foreign)
          (rename (chicken type) (: core-declare))
          (rename (crunch declarations) (: crunch-declare)))
  (import-for-syntax matchable
  					 (scheme base)
                     (chicken plist)
                     (chicken condition)
                     (crunch compiler)
                     (crunch compiler type))

  (define-for-syntax crunch:options-list '())

  (define-syntax crunch-options
    (er-macro-transformer
      (lambda (x r c)
        (set! crunch:options-list (strip-syntax (cdr x)))
        '(##core#undefined))))

  (define-syntax :
    (er-macro-transformer
      (lambda (x r c)
        `(,(r 'cond-expand)
            (crunched (,(r 'crunch-declare) ,@(cdr x)))
            (,(r 'else) (,(r 'core-declare) ,@(cdr x)))))))

  (define-syntax crunch
    (er-macro-transformer
      (lambda (x r c)
        (define prelude "#define CRUNCH_EMBEDDED")
        (define (emit-wrappers r emit)
          (lambda (mod exps)
            (emit
              (if mod
                (list (r 'begin))
                `(, (r 'begin)
                    ,@(map (match-lambda
                             ((n rn t)
                               (if (procedure-type? t)
                                   `(define ,rn ,(foreign-call-wrapper t n r))
                                    '(##core#undefined))))
                        exps))))))
        (let ((body (strip-syntax (cdr x))))
          (condition-case
            (let ((out '())
                  (code (open-output-string)))
             (apply crunch
             	    'export-all
                    `(consumer ,(lambda (c)
                                  (unless (eof-object? c)
                                    (display c code))))
                    `(export-consumer
                       ,(emit-wrappers r (lambda (x) (set! out (cons x out)))))
                    `(begin ,@body)
                    crunch:options-list)
            `(,(r 'begin)
               (,(r 'foreign-declare) ,prelude ,(get-output-string code))
               ,@out))
            ((crunch) (error 'crunch "compilation failed")))))))

  (foreign-declare "#define CRUNCH_EMBEDDED\n#include \"crunch.h\"")

  (define (string-result ptr)
    (let ((r (##sys#peek-nonnull-c-string ptr 0)))
      (##core#inline "C_crunch_unref" ptr)
      r))

  (define (u8vector-result ptr)
    (##core#inline "C_crunch_numvector_result" ptr))

  (define (s8vector-result ptr)
    (##sys#make-structure 's8vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (u16vector-result ptr)
    (##sys#make-structure 'u16vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (s16vector-result ptr)
    (##sys#make-structure 's16vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (u32vector-result ptr)
    (##sys#make-structure 'u32vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (s32vector-result ptr)
    (##sys#make-structure 's32vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (u64vector-result ptr)
    (##sys#make-structure 'u64vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (s64vector-result ptr)
    (##sys#make-structure 's64vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (f32vector-result ptr)
    (##sys#make-structure 'f32vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (f64vector-result ptr)
    (##sys#make-structure 'f64vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (c64vector-result ptr)
    (##sys#make-structure 'c64vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (c128vector-result ptr)
    (##sys#make-structure 'c128vector
      (##core#inline "C_crunch_numvector_result" ptr)))

  (define (intvector-result ptr)
    (##sys#make-structure
      (cond-expand (64bit 's64vector) (else 's32vector))
      (##core#inline "C_crunch_numvector_result" ptr)))

)