;;;; forth.scm

(import (scheme base)
        (scheme file)
        (chicken number-vector)
        (chicken memory)
        (scheme process-context)
        (chicken fixnum)
        mos6502)

(define cpu (make-processor))
(define mem (processor-memory cpu))
(define regs (processor-registers cpu))

(define stop #f)

(define TCONTROL_REGISTER #xc000)
(define TINPUT_CHAR #xc001)
(define TOUTPUT_CHAR #xc002)
(define TINPUT_CHAR_BIT #x80)
(define TOUTPUT_CHAR_BIT #x40)

(define stdin (open-input-string ": sum ( n - ) 0 swap 0 do i + loop ;\n10000 sum . cr\n"))
(define stdout (current-output-port))

(define (poll-terminal-device p)
  (let* ((mem (processor-memory p))
         (cr (u8vector-ref mem TCONTROL_REGISTER)))
    (cond ((zero? (fxand cr TINPUT_CHAR_BIT))
           (let ((c (read-char stdin)))
             (when (eof-object? c)
               (display "EOF from terminal\n" (current-error-port))
               (stop))
             (if (char=? c #\newline)
                 (set! c #\return)
                 (set! c (char-upcase c)))
             (u8vector-set! mem TINPUT_CHAR (fxand #x7f (char->integer c)))))
          ((zero? (fxand cr TOUTPUT_CHAR_BIT))
           (write-char (integer->char (fxand #x7f (u8vector-ref mem TOUTPUT_CHAR))) stdout)
           (flush-output stdout)))
    (u8vector-set! mem TCONTROL_REGISTER
      (fxior cr (fxior TINPUT_CHAR_BIT TOUTPUT_CHAR_BIT)))))

(define (read-all fname)
  (call-with-port (open-binary-input-file fname)
    (lambda (p)
      (read-bytevector #x10000 p))))

(define (load-bin fname addr)
  (let ((bin (read-all
  			   (let ((p (get-environment-variable "CRUNCH_TEST_DIR")))
  			   	 (if p
  			   	 	 (string-append p "/" fname)
  			   	 	 fname)))))
    (move-memory! bin mem (bytevector-length bin) 0 addr)))

(load-bin "6502/forth.bin" #x200)
(load-bin "6502/bios.bin" #xd000)

(reset! cpu)

(u8vector-set! regs 5 0)
(u8vector-set! regs 6 2)

(call/cc
  (lambda (k)
    (set! stop (lambda _ (k #f)))
    (do () (#f)
      (execute! cpu cycles: 10)
      (poll-terminal-device cpu))))
